

#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/sem.h>
#include <stdio.h>

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"
#include "http_request.h"
#include "http_main.h"
#include "http_conf_globals.h"	// ap_user_id
#include "dns_lock.h"






// return:	semid

int moddns_mutex_init ()
{
  int sem;
  struct semid_ds perm = { };

  perm.sem_perm.uid = ap_user_id;
  perm.sem_perm.gid = -1;
  perm.sem_perm.mode = 0600;

  sem = semget (IPC_PRIVATE, 1, IPC_CREAT|IPC_EXCL);
  if (sem < 0) {
    perror ("mod_dns: Failed to create mutex: semget");
    exit(1);
  }
  if (semctl (sem, 0, IPC_SET, &perm) < 0) {
    perror ("mod_dns: Failed to set mutex permissions: semctl");
    exit(1);
  }
  if (semctl (sem, 0, SETVAL, 1) < 0) {
    perror ("mod_dns: Failed to open mutex: semctl");
    exit(1);
  }
//fprintf (stderr, "created sem %d (pid=%d)\n", sem, getpid());
  return sem;
}


// remove mutex

void moddns_mutex_zap (int lock)
{
//fprintf(stderr, "zapping sem %d (pid=%d)\n", lock, getpid());
  if (semctl (lock, 0, IPC_RMID, 0) < 0) {
    perror ("mod_dns: Failed to remove mutex: semctl");
  }
}



// lock mutex

void moddns_mutex_lock (int lock)
{
  static struct sembuf op = { 0, -1, SEM_UNDO };
//  op.sem_num	= 0;
//  op.sem_op	= -1;
//  op.sem_flg	= SEM_UNDO;
  if (semop (lock, &op, 1) < 0) {
    perror ("mod_dns: Failed to aquire mutex: semop");
  }
}


// unlock mutex

void moddns_mutex_unlock (int lock)
{
  static struct sembuf op = { 0, 1, SEM_UNDO };
//  op.sem_num	= 0;
//  op.sem_op	= 1;
//  op.sem_flg	= SEM_UNDO;
  if (semop (lock, &op, 1) < 0) {
    perror ("mod_dns: Failed to release mutex: semop");
  }
}









