
#include <string.h>
#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_protocol.h"
#include "http_request.h"
#include "http_main.h"
#include "http_conf_globals.h"
#include "ap_mm.h"

#include "mod_dns.h"
#include "dns_lock.h"
#include "hash.h"
// djbdns
#include "buffer.h"
#include "exit.h"
#include "strerr.h"
#include "ip4.h"
#include "dns.h"
#include "stralloc.h"
#include "uint16.h"
#include "byte.h"
// end djbdns
 
module MODULE_VAR_EXPORT dns_module;



//BEGIN DEBUG
#define RD_Lock (shared->rd_Lock)
#define WR_Lock (shared->wr_Lock)
#define CHECK_RLOCK_IN	//{ if(WR_Lock) fprintf(stderr,"CHECK_RLOCK_IN: FUCK pid=%d!!!\n",getpid()); RD_Lock++; }
#define CHECK_RLOCK_OUT	//{ if(WR_Lock||RD_Lock<1) fprintf(stderr,"CHECK_RLOCK_OUT: FUCK w=%d,r=%d,pid=%d!!!\n",WR_Lock,RD_Lock,getpid()); RD_Lock--; }
#define CHECK_WLOCK_IN 	//{ if(WR_Lock) fprintf(stderr,"CHECK_WLOCK_IN: FUCK pid=%d!!!\n",getpid()); WR_Lock++; }
#define CHECK_WLOCK_OUT	//{ if(!WR_Lock||RD_Lock) fprintf(stderr,"CHECK_WLOCK_OUT: FUCK w=%d,r=%d,pid=%d!!!\n",WR_Lock,RD_Lock,getpid()); WR_Lock--; }
//END DEBUG



// locks: the ap_mm_lock/ap_mm_unlock functions are not reliable (on linux/i386 at least)
//#define SHARED_RD_LOCK()	ap_mm_lock (mm, AP_MM_LOCK_RD)
//#define SHARED_RW_LOCK()	ap_mm_lock (mm, AP_MM_LOCK_RW)
//#define SHARED_UNLOCK()	ap_mm_unlock (mm)
#define SHARED_RD_LOCK()	moddns_mutex_lock (shared->lock)
#define SHARED_RW_LOCK()	moddns_mutex_lock (shared->lock)
#define SHARED_UNLOCK() moddns_mutex_unlock (shared->lock)


#define OPT_INT_UNSET		(-1)
#define NVL_INT(this,that) ((this) != OPT_INT_UNSET ? (this) : (that))


typedef struct {
  int lookup_enabled;
  int timeout_ms;
  int resendtimer_ms;
} moddns_dir_config;

static void *moddns_create_dir_config(pool *p, char *path)
{
  moddns_dir_config *cfg = (moddns_dir_config *)ap_pcalloc(p, sizeof(moddns_dir_config));
  cfg->lookup_enabled	= OPT_INT_UNSET;
  cfg->timeout_ms	= atoi(OPT_DNS_TIMEOUT_DEFAULT);
  cfg->resendtimer_ms	= atoi(OPT_DNS_RESENDTIMER_DEFAULT);
  return (void *)cfg;
}


static void *moddns_merge_dir_config (pool *p, void *basev, void *addv)
{
  moddns_dir_config *new	= (moddns_dir_config *)ap_pcalloc (p, sizeof(moddns_dir_config));
  moddns_dir_config *base	= (moddns_dir_config *)basev;
  moddns_dir_config *add	= (moddns_dir_config *)addv;
    
  new->lookup_enabled	= NVL_INT (add->lookup_enabled, base->lookup_enabled);
  new->timeout_ms	= NVL_INT (add->timeout_ms, base->timeout_ms);
  new->resendtimer_ms	= NVL_INT (add->resendtimer_ms, base->resendtimer_ms);

  return (void*)new; 
}


static const char *moddns_cmd_AdnsLookups (cmd_parms *p, void *d, char *arg) 
{
  moddns_dir_config *cf = (moddns_dir_config *)d;
  if (!strcasecmp(arg, "on")) {
    cf->lookup_enabled	= OPT_DNS_LOOKUP_ON;
  } else if (!strcasecmp(arg, "off")) {
    cf->lookup_enabled	= OPT_DNS_LOOKUP_OFF;
  } else {
    return "parameter must be 'on' or 'off'";
  }
  return NULL;
}

static const char *moddns_cmd_AdnsTimeout_ms (cmd_parms *p, void *d, char *arg)
{
  moddns_dir_config *cf = (moddns_dir_config *)d;
  int timeout = atoi(arg);
  if (timeout > 0) {
    cf->timeout_ms	= timeout;
  } else {
    return "parameter must be positive number of milliseconds";
  }
  return NULL;
}

static const char *moddns_cmd_AdnsResendTimer_ms (cmd_parms *p, void *d, char *arg)
{
  moddns_dir_config *cf = (moddns_dir_config *)d;
  int timeout = atoi(arg);
  if (timeout > 0) {
    cf->resendtimer_ms	= timeout;
  } else {
    return "parameter must be positive number of milliseconds";
  }
  return NULL;
}




static command_rec moddns_cmds[] =
{
  {
	"DnsLookups",			/* directive name */
	moddns_cmd_AdnsLookups,		/* config action routine */
	NULL,				/* argument to include in call */
	OR_ALL,				/* where available */
	TAKE1,				/* arguments */
	"DNS lookup {on|off} switch; default is 'off'"
  },
  {
	"DnsTimeout_ms",		/* directive name */
	moddns_cmd_AdnsTimeout_ms,	/* config action routine */
	NULL,				/* argument to include in call */
	OR_ALL,				/* where available */
	TAKE1,				/* arguments */
	"DNS lookup timeout; default is '" OPT_DNS_TIMEOUT_DEFAULT "' ms"
  },
  {
	"DnsResendTimer_ms",		/* directive name */
	moddns_cmd_AdnsResendTimer_ms,	/* config action routine */
	NULL,				/* argument to include in call */
	OR_ALL,				/* where available */
	TAKE1,				/* arguments */
	"DNS UDP resend timeout; default is '" OPT_DNS_RESENDTIMER_DEFAULT "' ms"
  },
  {NULL}
};



typedef struct {
int rd_Lock;	// DEBUG
int wr_Lock;	// DEBUG
  // 
  hash_t	iphash;
  hnode_t**	htable;		// free at module exit
  struct Stats_t {
    long	hits;		// cache performance stats
    long	lookups;	// incremented for every external lookup attempt
    long	good;		// hostname found
    long	timeouts;	// hostname not found as a result of a timeout
    long	retries[5];	// number of UDP packets resends that eventually obtained a response
    double	timewait;	// how many seconds have we spent waiting for a dns reply?
  } stats;
  //
  int		allocated;	// shared memory size
  int		lock;		// semaphore
  int		capacity;
  int		rq_pos;		// write-pos in ring_queue
  struct {
    time_t	age;
    hnode_t*	node;
  } ring_queue[1];		// each time a new node is allocated, write its pointer to [rq_pos++]
} moddns_shared_t;

#define RETRY_HISTOGRAM_ARRAY_SIZE (sizeof(shared->stats.retries)/sizeof(shared->stats.retries[0]))





typedef struct {
  int		timeout_ms;	// total time for each query
  int		retry_ms;	// UDP resend timer
} Timeouts_t;









// module private globals
static moddns_shared_t *shared;
static AP_MM *mm;


// comparison function for searching ip hash
static int iphash_comp_f (const void *a, const void *b)
{
  return strcmp ((const char*)a, (const char*)b);
}
static hnode_t* moddns_hnode_alloc_f (void *mm)
{ 
  return ap_mm_malloc ((AP_MM*)mm, sizeof(hnode_t));
}
static void moddns_hnode_free_f (hnode_t* node, void *mm)
{
  ap_mm_free ((AP_MM*)mm, node);
}







////////////////////////////////////
// MODULE INITIALIZATION FUNCTION //
////////////////////////////////////


// cleanup

static void moddns_exit_hook (void* server)
{
  hscan_t it;
  hnode_t* node;
  int i;

//fprintf(stderr, "moddns_exit_hook: pid=%d\n", getpid());

  SHARED_RW_LOCK();
  {
    int ctotal = shared->stats.hits + shared->stats.lookups;

    errno = 0;
    ap_log_error (APLOG_MARK, APLOG_NOTICE, server, "mod_dns: Cache performance: %ld hits; %ld misses; %.02f%% hit ratio",
	shared->stats.hits,
	shared->stats.lookups,
	(ctotal ? 100.0*shared->stats.hits/ctotal : 0.0)
    );

    ap_log_error (APLOG_MARK, APLOG_NOTICE, server, "mod_dns: DNS performance: %ld good; %ld bad; %ld timeout; %.02f%% good; %.02fs avg",
	shared->stats.good,
	shared->stats.lookups - shared->stats.good,
	shared->stats.timeouts,
	(shared->stats.lookups ? 100.0*shared->stats.good/shared->stats.lookups : 0.0),
	(shared->stats.lookups ? shared->stats.timewait/shared->stats.lookups : 0.0)
    );


    // free all shared allocations..
    hash_scan_begin (&it, &shared->iphash);
    for (i=0; (node = hash_scan_next(&it)); ++i) {
      hash_scan_delete (&shared->iphash, node);
      ap_mm_free (mm, hnode_get(node));
      ap_mm_free (mm, (char*)hnode_getkey(node));
      moddns_hnode_free_f (node, mm);
    }
    errno = 0;
    ap_log_error (APLOG_MARK, APLOG_NOTICE, server, "mod_dns: Released %d/%d ip/host pairs from shared cache",
	i, shared->capacity);

    if (shared->htable) {
      ap_mm_free (mm, shared->htable);
      shared->htable = 0;
    }
  }
  SHARED_UNLOCK();
  moddns_mutex_zap (shared->lock);	// get rid of mutex

  ap_mm_free (mm, shared);
  shared = 0;

  ap_mm_destroy (mm);
  mm = 0;

}



// init

static void moddns_init_hook (server_rec *s, pool *p)
{
  int avail;

  // register cleanup

  ap_register_cleanup (p, s, moddns_exit_hook, ap_null_cleanup);

//fprintf(stderr, "moddns_init_hook: pid=%d\n", getpid());




  mm = ap_mm_create(DEFAULT_DNS_SHARED_SIZE, DEFAULT_DNS_SHARED_FILE);
  if (!mm) {
    if (ap_mm_error()) {
      fprintf (stderr, "mod_dns: Cannot allocate shared memory: %s\n", ap_mm_error());
    } else {
      fprintf (stderr, "mod_dns: Apache must be compiled with EAPI and mm support\n");
    }
    exit(1);
  }


  if (ap_mm_permission (mm, DNS_MM_FILE_MODE, ap_user_id, -1)) {
    perror ("mod_dns: ap_mm_permission");
    exit(1);
  }

  
  avail = ap_mm_available (mm);
  errno = 0;
  ap_log_error (APLOG_MARK, APLOG_NOTICE, s, "mod_dns: Allocated %d bytes of shared cache memory", avail);
//ap_mm_calloc (mm, 1, 7790);


  // yeah, we'll crash and burn if something fails, but at least an error is printed first..

  shared = ap_mm_calloc (mm, 1, sizeof(moddns_shared_t) + DEFAULT_DNS_HASH_SIZE * sizeof(shared->ring_queue[0]));
  if (!shared) {
    perror ("mod_dns: Failed to allocate shared fixed space");
    exit(1);
  }
  shared->capacity = DEFAULT_DNS_HASH_SIZE;
  shared->lock = moddns_mutex_init ();
  shared->allocated = avail;
  SHARED_RW_LOCK();

  shared->htable = ap_mm_malloc (mm, sizeof(hnode_t*) * DEFAULT_DNS_HASH_SIZE);
  if (!shared->htable) {
    perror ("mod_dns: Failed to allocated shared hash table");
    exit(1);
  }
  if (!hash_init (&shared->iphash, DEFAULT_DNS_HASH_SIZE, iphash_comp_f, NULL, shared->htable, DEFAULT_DNS_HASH_SIZE)) {
    perror ("mod_dns: Failed to initialize shared hash");
    exit(1);
  }
  errno = 0;
  ap_log_error (APLOG_MARK, APLOG_NOTICE, s, "mod_dns: Initialized shared cache with %d ip/host slots", DEFAULT_DNS_HASH_SIZE);

  // set node allocator to use shared memory
  hash_set_allocator (&shared->iphash, moddns_hnode_alloc_f, moddns_hnode_free_f, mm);


  SHARED_UNLOCK();
}


/*
  SHARED_RD_LOCK();
  SHARED_RW_LOCK();
  SHARED_UNLOCK();
*/





/////////////////////////////////////////////////////////////////
// success:	return 0
// failure:	return error code (EAGAIN for timeout)


static int dns_resolve_timed (const char *q, const char qtype[2], Timeouts_t* times)
{
  struct taia start;
  struct taia now;
  struct taia diff = { {times->timeout_ms/1000}, 1000000*(times->timeout_ms%1000), 0 };
  struct taia retry = { {times->retry_ms/1000}, 1000000*(times->retry_ms%1000), 0 };
  struct taia deadline = { { 0 } };
  struct taia chkpnt = { { 0 } };
  char servers[64];
  iopause_fd x[1];
  int r = 0;
  int transmits = 0;

  taia_now (&start);
  taia_add (&deadline, &start, &diff);	// final deadline

  if (dns_resolvconfip(servers) == -1) return -1;
  if (dns_transmit_start(&dns_resolve_tx,servers,1,q,qtype,"\0\0\0\0") == -1) return -1;

  while (r == 0) {
    taia_now (&now);
    chkpnt = retry;		// retry timeout
    taia_add (&chkpnt, &chkpnt, &now);
    if (taia_less(&deadline,&chkpnt)) {
      chkpnt = deadline;
    }
    if (taia_less(&chkpnt,&now)) {
      r = EAGAIN;		// timeout
      break;
    }
    transmits++;
    dns_transmit_io (&dns_resolve_tx, x, &chkpnt);
    iopause (x, 1, &chkpnt, &now);
    r = dns_transmit_get (&dns_resolve_tx, x, &now);
    if (r == -1) break;
    if (r == 1) { r = 0; break; }	// success
  }
  taia_now (&now);
  taia_sub (&diff, &now, &start);
  shared->stats.timewait += diff.sec.x + diff.nano / 1000000000.0;
  if (r != EAGAIN && transmits > 0) {
    if (transmits > RETRY_HISTOGRAM_ARRAY_SIZE) transmits = RETRY_HISTOGRAM_ARRAY_SIZE;
    shared->stats.retries[transmits-1] ++;
  }
  return r;
}


// unlike DJB's version, this routine fills in the allocated <q> pointer
// dns_name.c not used anymore

static int dns_name_packetp (stralloc *out,const char *buf,unsigned int len, char** qp)
{
  unsigned int pos;
  char header[12];
  uint16 numanswers;
  uint16 datalen;

  if (!stralloc_copys(out,"")) return -1;

  pos = dns_packet_copy(buf,len,0,header,12); if (!pos) return -1;
  uint16_unpack_big(header + 6,&numanswers);
  pos = dns_packet_skipname(buf,len,pos); if (!pos) return -1;
  pos += 4;

  while (numanswers--) {
    pos = dns_packet_skipname(buf,len,pos); if (!pos) return -1;
    pos = dns_packet_copy(buf,len,pos,header,10); if (!pos) return -1;
    uint16_unpack_big(header + 8,&datalen);
    if (byte_equal(header,2,DNS_T_PTR))
      if (byte_equal(header + 2,2,DNS_C_IN)) {
        if (!dns_packet_getname(buf,len,pos,qp)) return -1;
        if (!dns_domain_todot_cat(out,*qp)) return -1;
        return 0;
      }
    pos += datalen;
  }

  return 0;
}



//static char *q = 0;	// i wish djb didn't like globals so much... lame..

static int dns_name4_timed (stralloc *out, const char ip[4], Timeouts_t* times)
{
  int err = -1;
  char name[DNS_NAME4_DOMAIN];
  char *q = 0;

  dns_name4_domain(name,ip);
  err = dns_resolve_timed (name, DNS_T_PTR, times);
  if (err) return err;
  err = dns_name_packetp (out, dns_resolve_tx.packet, dns_resolve_tx.packetlen, &q);
  if (err == -1) return err;
  dns_transmit_free(&dns_resolve_tx);
  dns_domain_free(&q);
  return 0;
}


static int dns_ptr_lookup_djb (struct in_addr* addr_in, char *hostname_out, Timeouts_t* times)
{
  int err = -1;
  static stralloc out = { 0 };	// WARNING! Single threaded! keep using same buffer
  union {
    unsigned long int s_addr;
    char b[4];
  } ip;

  ip.s_addr = addr_in->s_addr;		// network byte-order

  err = dns_name4_timed (&out, ip.b, times);

  memcpy (hostname_out, out.s, out.len);
  hostname_out[out.len] = 0;

  return err;
}




static int dns_ptr_lookup (moddns_dir_config *cf, struct in_addr* addr_in, char *hostname_out)
{
  Timeouts_t	times;
  times.timeout_ms	= cf->timeout_ms;
  times.retry_ms	= cf->resendtimer_ms;
  return dns_ptr_lookup_djb (addr_in, hostname_out, &times);
}






static int moddns_lookup_hook (request_rec *r)
{
  moddns_dir_config *cf = MODDNS_DIR_CONFIG(r);
  conn_rec *cn = r->connection;
  int err;
  hnode_t *cnode;
  char* shm_addr = 0;
  char* shm_host = 0;
  int tries;
  char hostname[256] = "";

  if (cn->remote_host != NULL) {
    return OK;
  }
  if (cf->lookup_enabled != OPT_DNS_LOOKUP_ON) {
    return OK;		// lookup disabled..
  }

  SHARED_RD_LOCK();
CHECK_RLOCK_IN
//fprintf(stderr, "AP_MM_LOCK_RD locked\n");
  {
    cnode = hash_lookup (&shared->iphash, cn->remote_ip);
    if (cnode) {
      cn->remote_host = ap_pstrdup (cn->pool, (char*)hnode_get(cnode));
      shared->stats.hits++;
CHECK_RLOCK_OUT
      SHARED_UNLOCK();
      errno = 0;
//      ap_log_error (APLOG_MARK, APLOG_DEBUG, r->server, "mod_dns: Cached [%s] -> [%s]", cn->remote_ip, cn->remote_host);
      return OK;
    }
  }
CHECK_RLOCK_OUT
  SHARED_UNLOCK();

  shared->stats.lookups++;

  // perform IP -> HOST lookup

  err = dns_ptr_lookup (cf, &cn->remote_addr.sin_addr, hostname);

  errno = 0;	// clear so ap_log_error doesn't print junk

  if (err == EAGAIN) {
    // timeout
    shared->stats.timeouts++;
    ap_log_error (APLOG_MARK, APLOG_INFO, r->server, "mod_dns: Failed to resolve [%s]: %d ms timeout", cn->remote_ip, cf->timeout_ms);
  }
  else if (err) {
    ap_log_error (APLOG_MARK, APLOG_NOTICE, r->server, "mod_dns: Failed to resolve [%s]: %s", cn->remote_ip, strerror(err));
  }
  else {
    if (hostname[0]) {
      shared->stats.good++;
      cn->remote_host = ap_pstrdup (cn->pool, hostname);
      ap_str_tolower (cn->remote_host);
      ap_log_error (APLOG_MARK, APLOG_DEBUG, r->server, "mod_dns: Resolved [%s] -> [%s]", cn->remote_ip, cn->remote_host);
    } else {
//      ap_log_error (APLOG_MARK, APLOG_INFO, r->server, "mod_dns: Failed to resolve [%s]: %s", cn->remote_ip, moddns_strerror(answer->status));
    }
  }

  // if it failed, store "" in remote_host so we dont try it again..
  if (!cn->remote_host) cn->remote_host = "";

  SHARED_RW_LOCK();		// upgrade readlock to RW (well, for now we only use one type of lock..)
CHECK_WLOCK_IN
  {
//fprintf(stderr, "AP_MM_LOCK_RW locked\n");
    cnode = hash_lookup (&shared->iphash, cn->remote_ip);
    if (cnode) {
      // oops, someone else got here first..
      char* cachedhost = hnode_get(cnode);
      if (cachedhost[0] || !cn->remote_host[0]) {
	// ..and they cached a successful lookup, OR we got a failure too...
CHECK_WLOCK_OUT
	SHARED_UNLOCK();
	return OK;
      }
      hash_delete (&shared->iphash, cnode);	// unlink from list (since we reinsert below)
      ap_log_error (APLOG_MARK, APLOG_INFO, r->server, "mod_dns: Upgrading negative [%s] record with [%s]",
	cn->remote_ip, cn->remote_host);
      ap_mm_free (mm, cachedhost);
      shm_addr = (char*)hnode_getkey(cnode);	// reuse it below since it's the same IP!
    }


    // retrying not fully implemented: it currently doesn't try
    // to free up any memory even if allocation fails. see bottom of oop..
    for (tries=1; tries--; ) {

      if (hash_isfull(&shared->iphash)) {
	// buffer is full.	THIS CANNOT HAPPEN IF WE JUST CALLED hash_delete ABOVE
	hnode_t *trash = shared->ring_queue[shared->rq_pos].node;	// if this is null it means the code is fucked...
	if (!trash) {
	  ap_log_error (APLOG_MARK, APLOG_EMERG, r->server, "mod_dns: Shared memory corruption: node=%d", shared->rq_pos);
CHECK_WLOCK_OUT
	  SHARED_UNLOCK();	// try not to crash the server...
	}
	ap_log_error (APLOG_MARK, APLOG_DEBUG, r->server, "mod_dns: Recycled [%s] record: node=%d",
		(char*)hnode_getkey(trash), shared->rq_pos);
	hash_delete (&shared->iphash, trash);
	ap_mm_free (mm, (void*)hnode_getkey(trash));	// free up old key
	ap_mm_free (mm, hnode_get(trash));	// free up old value
	moddns_hnode_free_f (trash, mm);
	shared->ring_queue[shared->rq_pos].node = NULL;
      }

      if (!shm_addr) shm_addr = ap_mm_strdup (mm, cn->remote_ip);

      if (!shm_host) shm_host = ap_mm_strdup (mm, cn->remote_host);

      if (!cnode) {
	cnode = moddns_hnode_alloc_f (mm);
	if (cnode) hnode_init (cnode, 0);
      }

      if (shm_addr && shm_host && cnode) {
	hnode_put (cnode, shm_host);
	shared->ring_queue[shared->rq_pos].age = time(0);
	shared->ring_queue[shared->rq_pos].node = cnode;
	shared->rq_pos = (shared->rq_pos + 1) % shared->capacity;
	hash_insert (&shared->iphash, cnode, shm_addr);
	break;
      }
      // allocation failed, free up some space..
      errno = 0;
      ap_log_error (APLOG_MARK, APLOG_ERR, r->server, "mod_dns: Shared memory full: %ld/%d nodes cached",
		hash_count(&shared->iphash), shared->capacity);
      // try to free something here *TODO*
    }
    if (tries < 0) {
      ap_log_error (APLOG_MARK, APLOG_ERR, r->server, "mod_dns: Failed allocate shared cache memory: %s", ap_mm_error());
      if (shm_addr) ap_mm_free (mm, shm_addr);
      if (shm_host) ap_mm_free (mm, shm_host);
      if (cnode) moddns_hnode_free_f (cnode, mm);
    }
  }
CHECK_WLOCK_OUT
  SHARED_UNLOCK();

  return OK;
}




// display statistcs handler

static int moddns_stats_handler (request_rec *r)
{
  struct Stats_t *s = &shared->stats;

  r->content_type	= "text/html";
  ap_send_http_header (r);

  // no locking necessary for reading data.. (any inconsistencies are
  {
    int usage;
    time_t oldest;
    struct { int hh; int mm; int ss; } age = { 0 };
    long reqtotal = s->hits + s->lookups;
    long baddns = s->lookups - s->good;
    long responses = s->lookups - s->timeouts;
    int rq_pos = shared->rq_pos;

    if (shared->ring_queue[rq_pos].node) { usage = shared->capacity; }
    else { usage = rq_pos; }

    oldest = shared->ring_queue[rq_pos].age;
    if (!oldest && rq_pos>0) oldest = shared->ring_queue[0].age;
    if (oldest) age.ss = time(0) - oldest;
    age.mm = age.ss/60;		age.ss %= 60;
    age.hh = age.mm/60;		age.mm %= 60;

    ap_rprintf (r,
	"<head>\n"
	"  <title>mod_dns statistics [%s]</title>\n"
	"</head>\n"
	"<body>\n"
	"<table height=\"100%%\" width=\"100%%\" cellspacing=0 cellpadding=0 border=0><tr><td align=center valign=middle>\n"
	"<table border=0 cellpadding=4 cellspacing=2 bgcolor=white>\n"
	"  <tr align=right>	<td rowspan=4 align=left valign=top>Cache</td>\n"
	"				<td align=left>Usage/Size</td>	<td>%d/%d</td>	<td>%.02f%%</td>		</tr>\n"
	"  <tr align=right>		<td align=left>Age</td>		<td>%02d:%02d:%02d</td>	<td></td>		</tr>\n"
	"  <tr align=right>		<td align=left>Hits</td>	<td>%ld</td>	<td>%.02f%%</td>		</tr>\n"
	"  <tr align=right>		<td align=left>Misses</td>	<td>%ld</td>	<td>%.02f%%</td>		</tr>\n"
	"  <tr align=right>	<td rowspan=2 align=left valign=top>Shared memory</td>\n"
	"				<td align=left>Total</td>	<td>%d bytes</td>	<td></td>		</tr>\n"
	"  <tr align=right>		<td align=left>Free</td>	<td>%d bytes</td>	<td>%.02f%%</td>	</tr>\n"
	"  <tr align=right>	<td rowspan=5 align=left valign=top>DNS Resolver</td>\n"
	"				<td align=left>Total lookups</td>	<td>%ld</td>	<td></td>		</tr>\n"
	"  <tr align=right>		<td align=left>Good lookups</td>	<td>%ld</td>	<td>%.02f%%</td>	</tr>\n"
	"  <tr align=right>		<td align=left>Bad lookups</td>	<td>%ld</td>	<td>%.02f%%</td>		</tr>\n"
	"  <tr align=right>		<td align=left>Lookup timeouts</td>	<td>%ld</td>	<td>%.02f%%</td>	</tr>\n"
	"  <tr align=right>		<td align=left>Avg lookup time</td>	<td>%.02f ms</td>	<td></td>	</tr>\n"
	"  <tr align=right>	<td rowspan=5 align=left valign=top>Server responding</td>\n"
	"				<td align=left>Without retries</td>	<td>%ld</td>	<td>%.02f%%</td>	</tr>\n"
	"  <tr align=right>		<td align=left>After 1 retry</td>	<td>%ld</td>	<td>%.02f%%</td>	</tr>\n"
	"  <tr align=right>		<td align=left>After 2 retries</td>	<td>%ld</td>	<td>%.02f%%</td>	</tr>\n"
	"  <tr align=right>		<td align=left>After 3 retries</td>	<td>%ld</td>	<td>%.02f%%</td>	</tr>\n"
	"  <tr align=right>		<td align=left>After 4+ retries</td>	<td>%ld</td>	<td>%.02f%%</td>	</tr>\n"
	"</table>\n"
	"  <tr align=center valign=bottom>	<td colspan=4 height=1><font size=\"-2\">mod_dns v%s - Copyright &copy; 2001 Fredrik Sjoholm &lt;fredrik@sjoholm.com&gt; - <a href=\"http://oss.ezic.com/\">http://oss.ezic.com/</a></font></td>		</tr>\n"
	"</td></tr></table>\n"
	"</body>\n",
		ap_get_server_name (r),
		usage, shared->capacity,	100.0*usage/shared->capacity,
		age.hh, age.mm, age.ss,
		s->hits,		reqtotal ? 100.0*s->hits/reqtotal : 0.0,
		s->lookups,		reqtotal ? 100.0*s->lookups/reqtotal : 0.0,
			shared->allocated,
			ap_mm_available(mm),	100.0*ap_mm_available(mm)/shared->allocated,
		s->lookups,
		s->good,		s->lookups ? 100.0*s->good/s->lookups : 0.0,
		baddns,			s->lookups ? 100.0*baddns/s->lookups : 0.0,
		s->timeouts,		s->lookups ? 100.0*s->timeouts/s->lookups : 0.0,
		s->lookups ? 1000.0*s->timewait/s->lookups : 0.0,
		s->retries[0],		responses ? 100.0*s->retries[0]/responses : 0.0,
		s->retries[1],		responses ? 100.0*s->retries[1]/responses : 0.0,
		s->retries[2],		responses ? 100.0*s->retries[2]/responses : 0.0,
		s->retries[3],		responses ? 100.0*s->retries[3]/responses : 0.0,
		s->retries[4],		responses ? 100.0*s->retries[4]/responses : 0.0,
		MOD_DNS_VERSION
    );
  }

  return OK;
}



static handler_rec moddns_handlers[] =
{
  { "dns-stats-handler",		moddns_stats_handler },
  {NULL}
};
 


/* Tell Apache what phases of the transaction we handle */
module MODULE_VAR_EXPORT dns_module =
{
  STANDARD_MODULE_STUFF,
  moddns_init_hook,		/* module initializer			*/
  moddns_create_dir_config,	/* per-directory config creator		*/
  moddns_merge_dir_config,	/* dir config merger			*/
  NULL,				/* server config creator		*/
  NULL,				/* server config merger			*/
  moddns_cmds,			/* command table			*/
  moddns_handlers,		/* [7]  content handlers		*/
  NULL,				/* [2]  URI-to-filename translation	*/
  NULL,				/* [5]  check/validate user_id		*/
  NULL,				/* [6]  check user_id is valid *here*	*/
  NULL,				/* [4]  check access by host address	*/
  NULL,				/* [7]  MIME type checker/setter	*/
  NULL,				/* [8]  fixups				*/
  NULL,				/* [9]  logger				*/
  moddns_lookup_hook,		/* [3]  header parser			*/
  NULL,				/* process initialization		*/
  NULL,				/* process exit/cleanup			*/
  NULL				/* [1]  post read_request handling	*/
};







